"""
All manipulations
-----------------
"""

# Built-in modules #
import sys, types, pkgutil, sqlite3

# Internal modules #
import track
from track import manips, Track, FeatureStream
from track.extras import TrackCollection, VirtualTrack
from track.common import import_module, collapse, andify_strings

# Constants #
base_attributes = ['label', 'short_name', 'long_name', 'input_tracks', 'input_args',
                   'input_meta', 'output_tracks', 'tracks_collapse', 'chroms_collapse',
                   'fields_collapse', 'tooltip', 'numeric_example', 'visual_example',
                   'tests']
base_functions = ['generate']

# Conveniance functions #
is_gen   = lambda x: isinstance(x, sqlite3.Cursor) or \
                     isinstance(x, FeatureStream) or \
                     isinstance(x, types.GeneratorType) or \
                     isinstance(x, type(iter(()))) or \
                     isinstance(x, type(iter([])))
is_list  = lambda x: isinstance(x, tuple) or \
                     isinstance(x, list)
is_track = lambda x: isinstance(x, Track) or \
                     isinstance(x, TrackCollection) or \
                     isinstance(x, VirtualTrack)

################################################################################
class Manipulation(object):
    def __init__(self, *args):
        """Taking the module object of a standard manipulation file
            will build a callable instance satisfying the three
            access methods:

                1) overlap('tracks/pol2.sql', 'tracks/rap1.sql') # returns a virtual track
                2) overlap(pol2,rap1) # returns a virtual track
                3) overlap(pol2.read(chrom), rap1.read(chrom)) # returns a generator

            Other special cases:

                1) window_smoothing(pol2,rap1,L=100)
                2) merge_scores([pol2,rap1,ifhl,fhl1])
                3) overlap(pol2,rap1,ifhl,fhl1)
                4) complement(pol2.read(chrom), l=pol2.chrmeta[chrom]['length'])
        """
        # Only one argument #
        self.module = args[0]
        # Module attributes become instance attributes #
        for attr in base_attributes + base_functions: setattr(self, attr, getattr(self.module,attr))
        # Update the documentation #
        self.__doc__ = self.documentation

    @property
    def documentation(self):
        """Generate the sphinx docstring from the attributes of the manipulation."""
        # The long title #
        msg = self.long_name + "."
        # The tool tip #
        msg += "\n\n" + self.tooltip + '\n'
        # The input tracks #
        for t in self.input_tracks:
            if t.get('kind') == 'many':
                sub_msg = "\n:param %s: An arbitrary number of tracks or paths to tracks." \
                          " Eventually, generators yielding features."
            else:
                sub_msg = "\n:param %s: A track or the path to a track. Eventually, a generator yielding features."
            sub_msg = sub_msg % t['key']
            if 'fields' in t:
                if t.get('kind')=='many': sub_msg += " The fields read form these tracks will be: ``%s``"
                else:                     sub_msg += " The fields read form this track will be: ``%s``"
                sub_msg = sub_msg % t['fields']
                if '...' in t['fields']: sub_msg += ", extra fields will be used."
                else:                    sub_msg += ", extra fields will not be used."
            if t.get('no-overlap'): sub_msg += " The track is assumed to have no overlapping features."
            msg += sub_msg + '\n'
            if t.get('kind') == 'many': msg += ":type %s: list\n" % t['key']
        # The arguments #
        for p in self.input_args:
            msg += ":param %s:" % p['key']
            if 'doc' in p: msg += " " + p['doc']
            if 'default' in p: msg += " By default ``%s``." % p['default']
            if 'type' in p: msg += "\n:type %s: %s" % (p['key'], p['type'].__name__)
            msg += '\n'
        # The special parameter #
        for p in self.input_meta:
            msg += "\n:param %s:" % p['key']
            if 'kind' in p and p['kind'] == 'chrom_len':
                msg += " The length of the current chromosome " \
                       "(only necessary when calling the manipulation with generators).\n" \
                       ":type %s: int" % p['key']
            msg += '\n'
        # The output #
        for t in self.output_tracks:
            if 'fields' in t: msg += ":returns: A virtual track with the following fields: ``%s``." % t['fields']
            else:             msg += ":returns: A virtual track."
            msg += '\n'
        # The numeric example #
        if self.numeric_example:
            msg += '\nA numerical example::\n %s\n' % self.numeric_example.replace('\n','\n     ')
        # The viusal example #
        if self.visual_example:
            msg += '\nA visual example::\n %s\n' % self.visual_example.replace('\n','\n    ')
        # Chromosomes collapse #
        if self.chroms_collapse: msg += "\n"\
        "If the list of chromosomes contained in the various tracks differ," \
        " the conflict will be resolved by applying the '%s' principle." \
        % self.chroms_collapse
        # Tracks collapse #
        if self.tracks_collapse: msg += "\n"\
        "If the list of tracks supplied is more than the two authorized," \
        " the conflict will be resolved by applying the '%s' principle." \
        % self.tracks_collapse
        # Return the message #
        return msg

    def test(self):
        """Runs all unittest. Yields *got* and *expected*"""
        for test in self.tests:
            # Call the manipulation #
            final_args = {}
            for d in (test.get('tracks',{}), test.get('args',{})): final_args.update(d)
            result = self(**final_args)
            # In case we have a generator #
            if is_gen(result): got = list(result)
            # In case we have a track #
            if is_track(result):
                if 'chromosome' in test: got = list(result.read(test['chromosome']))
                else:                    got = list(result.read())
                result.close()
            # Return two elements #
            got = [tuple(x) for x in got]
            yield got, test['expected']

    def __call__(self, *args, **kwargs):
        """Check that all arguments are present
           and load all tracks that are given as paths
           instead of track objects. Also checks for
           direct calls with generators."""
        # Initialization #
        generator_call   = False # Special switch for direct generator calls
        found_args       = {}    # Will contain a set of parameters extracted
        found_tracks     = {}    # Will contain a set of track parameters extracted
        found_generators = {}    # Will contain a set of FeatureStream
        extra_args       = {}    # Will contain a set of parameters computed
        all_tracks       = []    # Will contain all single tracks sent
        tracks_to_close  = []    # Will contain single tracks to close
        virtual_tracks   = []    # Will contain the results tracks
        rest_of_fields   = []    # Will contain variable output fields when required
        ### Parse arguments ###
        for p in self.input_args:
            if p['key'] in kwargs: value = kwargs[p['key']]
            elif len(args) >= p['position']: value = args[p['position']-1]
            elif 'default' in p: value = p['default']
            elif p.get('optional'): continue
            else: raise Exception("The argument '%s' is missing for the manipulation '%s'." \
                                  % (p['key'], self.short_name))
            # Cast it if it's not the right type #
            if not isinstance(value, p['type']): value = p['type'](value)
            # Add it to the dict #
            found_args[p['key']] = value
        ### Parse tracks ###
        for t in self.input_tracks:
            if t['key'] in kwargs: value = kwargs[t['key']]
            elif len(args) >= t['position']: value = args[t['position']-1]
            elif 'default' in t: value = t['default']
            elif t.get('optional'): continue
            else: raise Exception("The argument '%s' is missing for the manipulation '%s'." \
                                  % (t['key'], self.short_name))
            # Check is track collection #
            if t.get('kind') == 'many':
                if not is_list(value):
                    message = "The track collection '%s' for the manipulation '%s' is not a list: %s"
                    raise Exception(message % (t['key'], self.short_name, value))
            # Don't modify the input list #
            if t.get('kind') == 'many': value = value[:]
            # Check for generator case #
            if is_gen(value): generator_call = True
            if t.get('kind') == 'many' and is_gen(value[0]): generator_call = True
            if generator_call:
                found_tracks[t['key']] = value
                continue
            # Check is path #
            if isinstance(value, basestring):
                value = track.load(value, readonly=True)
                tracks_to_close.append(value)
            if t.get('kind') == 'many':
                for i,_ in enumerate(value):
                    if isinstance(value[i], basestring):
                        value[i] = track.load(value[i], readonly=True)
                        tracks_to_close.append(value[i])
            # Add to the list of all tracks #
            if t.get('kind') == 'many': all_tracks += [x for x in value]
            else:                       all_tracks += [value]
            # Track collection must be combined #
            if t.get('kind') == 'many':
                value = TrackCollection(value, self.fields_collapse, self.chroms_collapse)
            # Variable fields case (track collection must collapse fields) #
            if t['fields'][-1] == '...':
                first_fields = t['fields'][:-1]
                rest_of_fields = [f for f in value.fields if f not in first_fields]
                value.fields = first_fields + rest_of_fields
            # Specific fields case #
            else: value.fields = t['fields']
            # What about track SimpleTrack case #
            pass #TODO
            # Add it to the dict #
            found_tracks[t['key']] = value
        # Check for generator case #
        if generator_call: return self.from_generator(found_tracks, found_args, args, kwargs)
        # Collapse chromosomes #
        if not self.chroms_collapse: chromosomes = all_tracks[0].chromosomes
        else: chromosomes = collapse(self.chroms_collapse, [t.chromosomes for t in all_tracks])
        # Multiple output tracks disabled #
        t = self.output_tracks[0]
        # Make a new virtual track #
        vtrack = VirtualTrack()
        # Output chromosome metadata #
        for chrom in chromosomes:
            vtrack.chrmeta[chrom] = {'length': max([i.chrmeta[chrom]['length'] for i in all_tracks])}
        # Output attributes #
        if t.get('datatype'): vtrack.datatype = t['datatype']
        # Output name #
        vtrack.name = self.long_name + ' on ' + andify_strings([i.name for i in all_tracks])
        ### Iterate on chromosomes ###
        for chrom in chromosomes:
            # Get special input arguments #
            for p in self.input_meta:
                if p['kind'] == 'chrom_len':
                    extra_args[p['key']] = vtrack.chrmeta[chrom]['length']
            # Call read on tracks #
            for k,input_track in found_tracks.items():
                if is_list(input_track): found_generators[k] = [i.read(chrom) for i in input_track]
                else:                    found_generators[k] = input_track.read(chrom)
            # What about track collapse and recursion #
            pass #TODO
            # Final argument list #
            final_args = {}
            for d in (found_args, found_generators, extra_args): final_args.update(d)
            # Call generate #
            data = self.generate(**final_args)
            # Variable fields case #
            if t['fields'][-1] == '...': fields = t['fields'][:-1] + rest_of_fields
            else:                        fields = t['fields']
            # Make a FeatureStream #
            stream = FeatureStream(data, fields)
            # Add it to the virtual track #
            vtrack.write(chrom, stream)
        # Close tracks later #
        vtrack.tracks_to_close = tracks_to_close
        # Add it #
        virtual_tracks.append(vtrack)
        # Return one virutal track or list of virtual tracks #
        return len(virtual_tracks) == 1 and virtual_tracks[0] or virtual_tracks

    def from_generator(self, found_generators, found_args, args, kwargs):
        """To be used when the manipulation is accessed
        directly with generators"""
        # Initialization #
        extra_args = {}
        # Parse special input arguments #
        for p in self.input_meta:
            if p['key'] in kwargs: value = kwargs[p['key']]
            elif len(args) >= p['position']: value = args[p['position']-1]
            elif p.get('default'): value = p['default']
            elif p.get('optional'): continue
            else: raise Exception("The special argument '%s' is missing for the generator '%s'." \
                                  % (p['kind'], self.short_name))
            # Cast it if it's not the right type #
            if not isinstance(value, p['type']): value = p['type'](value)
            # Add it to the dict #
            extra_args[p['key']] = value
        # Final argument list #
        final_args = {}
        for d in (found_args, found_generators, extra_args): final_args.update(d)
        return self.generate(**final_args)

################################################################################
# This module #
self_module = sys.modules[__name__]
# Where are the all the manipulations #
manips_directory = manips.__path__[0]
# A list containing their names #
list_of_manip_names = [name for loader, name, ispkg in pkgutil.iter_modules([manips_directory])]
# For every one make a callable function #
for manip_name in list_of_manip_names:
    manip_module = import_module('track.manips.' + manip_name)
    setattr(self_module, manip_name, Manipulation(manip_module))
# Add the documentation for every one #
self_module.__doc__ += '\n'.join(['.. autofunction:: %s' % s for s in list_of_manip_names])

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the parsing of bedgraph files according to this standard:

http://genome.ucsc.edu/goldenPath/help/bedgraph.html
"""

# Built-in modules #
import shlex

# Internal modules #
from track.parse import Parser
from track.common import iterate_lines

# Constants #
all_fields = ['start', 'end', 'score']

################################################################################
class ParserBedgraph(Parser):
    format = 'bedgraph'
    def parse(self):
        # Initial variables #
        info = {}
        declare_track = True
        # Main loop #
        for number, line in iterate_lines(self.path):
            # Ignored lines #
            if line.startswith("browser "): continue
            # Track headers #
            if line.startswith("track "):
                try:
                    info = dict([p.split('=',1) for p in shlex.split(line[6:])])
                except ValueError:
                    self.handler.error("The track%s seems to have an invalid <track> header line", self.path, number)
                declare_track = True
                continue
            # Split the lines #
            items = line.split('\t')
            if len(items) == 1: items = line.split()
            # Chromosome #
            chrom = items.pop(0)
            # Length is three #
            if len(items) != 3:
                self.handler.error("The track%s doesn't have four columns", self.path, number)
            # Have we started a track already ? #
            if declare_track:
                declare_track = False
                self.handler.defineFields(all_fields)
                self.handler.newTrack(info, self.name)
            # Start and end fields #
            try:
                items[0] = int(items[0])
                items[1] = int(items[1])
            except ValueError:
                self.handler.error("The track%s has non integers as interval bounds", self.path, number)
            # Score field #
            if items[2] == '.' or items[2] == '': items[2] = 0.0
            try:
                items[2] = float(items[2])
            except ValueError:
                self.handler.error("The track%s has non floats as score values", self.path, number)
            # Yield it #
            self.handler.newFeature(chrom, items)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains tests where the cursor of the track is used to execute
SQL statements out of the control of the library.
"""

# Internal modules #
import track
from track.common import temporary_path

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestNewTables(unittest.TestCase):
    def runTest(self):
        out_path = temporary_path('.sql')
        with track.new(out_path) as t:
            for chrom in range(5): t.write(str(chrom), [(0,10,'A',0.0,-1)])
            cur = t.cursor()
            cur.execute("CREATE table tmp (koopa text,troopa text)")
            cur.execute("INSERT into  tmp values (?,?)", (1,2))

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the sql serialization.
"""

# Built-in modules #
import time

# Internal modules #
import track
from track.serialize import Serializer
from track.common import make_file_names

# Constants #
BUFFER_SIZE = 32767

################################################################################
class SerializerSQL(Serializer):
    format = 'sql'

    def __enter__(self):
        self.buffer = []
        self.fields = None
        self.current_track = None
        self.current_chrom = None
        self.file_paths = make_file_names(self.path)
        return self

    def __exit__(self, errtype, value, traceback):
        if self.current_track: self.closeCurrentTrack()

    def defineFields(self, fields):
        if fields != self.fields: self.flushBuffer()
        self.fields = fields

    def defineChrmeta(self, chrmeta):
        self.current_chrmeta = chrmeta

    def defineAssembly(self, assembly):
        self.current_assembly = assembly

    def newTrack(self, info=None, name=None):
        # Close previous track #
        if self.current_track: self.closeCurrentTrack()
        # Get a file name #
        path = self.file_paths.next()
        # Add it to the result #
        self.tracks.append(path)
        # Create it #
        self.current_track = track.new(path)
        # Add the metadata #
        if info: self.current_track.info.update(info)
        # Add the tags #
        #if name: self.current_track.info['converted_from'] = name
        #self.current_track.info['converted_by'] = 'track package'
        #self.current_track.info['converted_at'] = time.asctime()
        # Benchmark it #
        #self.start_time = time.time()

    def newFeature(self, chrom, feature):
        if chrom == self.current_chrom and len(self.buffer) < BUFFER_SIZE:
            self.buffer.append(feature)
        else:
            self.flushBuffer()
            self.current_chrom = chrom
            self.buffer.append(feature)

    #-----------------------------------------------------------------------------#
    def closeCurrentTrack(self):
        # Empty buffer #
        self.flushBuffer()
        # Add chrmeta #
        if hasattr(self, 'current_chrmeta'): self.current_track.chrmeta = self.current_chrmeta
        # Add assembly #
        if hasattr(self, 'current_assembly'): self.current_track.assembly = self.current_assembly
        # Add the benchmark #
        #self.current_track.info['converted_in'] = time.time() - self.start_time
        # Commit changes #
        self.current_track.save()
        self.current_track.close()
        # Reset varaibles #
        self.current_track = None

    def flushBuffer(self):
        if self.buffer:
            self.current_track.write(self.current_chrom, self.buffer, self.fields)
            self.buffer = []

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains simple reading tests.
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.common import temporary_path
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestWithChrom(unittest.TestCase):
    """Read all features from a track"""
    def runTest(self):
        in_path = samples['small_features'][1]['sql']
        with track.load(in_path) as t:
            data = t.read()
            got = list(data)
        expected = [('chrI', 0, 10, u'Validation feature 1', 10.0),
                    ('chrI', 2, 8, u'Validation feature 2', 0.0),
                    ('chrI', 20, 30, u'Validation feature 3', 10.0),
                    ('chrI', 25, 30, u'Validation feature 4', 0.0),
                    ('chrI', 40, 45, u'Validation feature 5', 0.0),
                    ('chrI', 40, 50, u'Validation feature 6', 10.0),
                    ('chrI', 60, 70, u'Validation feature 7', 10.0),
                    ('chrI', 70, 80, u'Validation feature 8', 10.0),
                    ('chrI', 90, 100, u'Validation feature 9', 10.0),
                    ('chrI', 90, 110, u'Validation feature 10', 10.0),
                    ('chrI', 120, 130, u'Validation feature 11', 10.0),
                    ('chrI', 125, 135, u'Validation feature 12', 5.0)]
        self.assertEqual(got, expected)

class TestWithoutName(unittest.TestCase):
    """Read some features from a track"""
    def runTest(self):
        in_path = samples['small_features'][5]['sql']
        with track.load(in_path) as t:
            data = t.read('chrI', ('start', 'end', 'name', 'score'))
            got = tuple(data.next())
        expected = (14, 19, u'', 0.0)
        self.assertEqual(got, expected)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



from distutils.core import setup, Extension

setup(
        name             =   'track',
        version          =   '1.1.0',
        description      =   'Python package for reading and writing genomic data',
        long_description =   open('README.md').read(),
        license          =   'GNU General Public License 3.0',
        url              =   'http://bbcf.epfl.ch/track',
        author           =   'EPFL BBCF',
        author_email     =   'webmaster.bbcf@epfl.ch',
        classifiers      =   ['Topic :: Scientific/Engineering :: Bio-Informatics'],
        packages         =   ['track',
                              'track.parse',
                              'track.serialize',
                              'track.test',
                              'track.manips',
                             ],
        package_data     = {'track': ['genomes.db']},
        ext_modules      = [Extension('track.pyrow', ['src/pyrow.c'])],
        scripts          = ['track/track'],
    )



# coding: utf-8

################################### Modules ####################################
import sys
from track.test import samples

################################# Description ##################################
label           = 'neighborhood'
short_name      = 'Neighborhood regions'
long_name       = 'Compute neighborhood regions upstream and downstream of features.'
input_tracks    = [{'key':'X', 'position':1, 'kind':'single', 'fields':['start','end','...']}]
input_args      = [{'key':'before_start', 'position':3, 'type': int, 'default': 0,
                    'doc':'Base pairs before the feature start.'},
                   {'key':'after_end',    'position':4, 'type': int, 'default': 0,
                    'doc':'Base pairs after the feature end.'},
                   {'key':'after_start',  'position':5, 'type': int, 'default': 0,
                    'doc':'Base pairs after the feature start.'},
                   {'key':'before_end',   'position':6, 'type': int, 'default': 0,
                    'doc':'Base pairs before the feature end.'},
                   {'key':'on_strand', 'position':7, 'type': bool, 'default': False,
                    'doc':'Features on the negative strand can be inverted.'},]
input_meta      = [{'key':'l', 'position':2, 'type': int, 'kind':'chrom_len'}]
output_tracks   = [{'position':1, 'fields': ['start','end','...']}]
tracks_collapse = None
fields_collapse = None
chroms_collapse = None

################################ Documentation #################################
tooltip = \
"""
Given a stream of features and four integers ``before_start``, ``after_end``,
``after_start`` and ``before_end``, this manipulation will output,
for every feature in the input stream, one or two features
in the neighborhood of the original feature.

   * Only ``before_start`` and ``after_end`` are given::

         (start, end, ...) -> (start+before_start, end+after_end, ...)

   * Only ``before_start`` and ``after_start`` are given::

         (start, end, ...) -> (start+before_start, start+after_start, ...)

   * Only ``after_end`` and ``before_end`` are given::

         (start, end, ...) -> (end+before_end, end+after_end, ...)

   * If all four parameters are given, a pair of features is outputted::

         (start, end, ...) -> (start+before_start, start+after_start, ...)
                              (end+before_end, end+after_end, ...)

   * If the boolean parameter ``on_strand`` is set to True,
     features on the negative strand are inverted as such::

         (start, end, ...) -> (start-after_end, start-before_end, ...)
                              (end-after_start, end-before_start, ...)
"""

tickets = ['https://github.com/bbcf/gMiner/issues/1']

################################## Examples ####################################
numeric_example = \
"""
X [start,end]: (10,20)
R [start,end]: (5,8) (22,25)
"""

visual_example = \
u"""
X: ──────────▤▤▤▤▤▤▤▤▤──────────────
R: ────▤▤▤▤─────────────▤▤▤▤────────
"""

#################################### Tests #####################################
tests = [
    {'tracks':     {'X': samples['small_features'][1]['sql']},
     'chromosome': 'chrI',
     'args':       {'l': 135, 'before_start':-2, 'after_end':2},
     'expected':  [( 0,   12,  u'Validation feature 1',  10.0),
                   ( 0,   10,  u'Validation feature 2',   0.0),
                   (18,   32,  u'Validation feature 3',  10.0),
                   (23,   32,  u'Validation feature 4',   0.0),
                   (38,   47,  u'Validation feature 5',   0.0),
                   (38,   52,  u'Validation feature 6',  10.0),
                   (58,   72,  u'Validation feature 7',  10.0),
                   (68,   82,  u'Validation feature 8',  10.0),
                   (88,  102,  u'Validation feature 9',  10.0),
                   (88,  112,  u'Validation feature 10', 10.0),
                   (118, 132,  u'Validation feature 11', 10.0),
                   (123, 137,  u'Validation feature 12',  5.0)]}]

############################### Implementation #################################
def bounded(X, min_bound=0, max_bound=sys.maxint):
    for x in X:
        if not x[1] < min_bound and not x[0] > max_bound and x[0] < x[1]:
            yield (max(x[0],min_bound), min(x[1],max_bound)) + x[2:]

def neighborhood(X, l, before_start=0, after_end=0, after_start=0, before_end=0, on_strand=False):
    if before_start and after_start:
        if before_start > after_start:
            raise Exception("'before_start' cannot be larger than 'after_start'")
    if before_end and after_end:
        if before_end > after_end:
            raise Exception("before_end cannot be larger than 'after_end'")
    if not on_strand:
        if before_start and after_end and after_start and before_end:
            for x in X:
                yield (x[0]+before_start, x[0]+after_start) + x[2:]
                yield (x[1]+before_end,   x[1]+after_end) + x[2:]
        if before_start and after_start:
            for x in X: yield (x[0]+before_start, x[0]+after_start)  + x[2:]
        if before_end and after_end:
            for x in X: yield (x[1]+before_end,   x[1]+after_end)    + x[2:]
        if before_start and after_end:
            for x in X: yield (x[0]+before_start, x[1]+after_end)    + x[2:]
    else:
        if before_start and after_end and after_start and before_end:
            for x in X:
                yield (x[0]-after_end,   x[0]-before_end) + x[2:]
                yield (x[1]-after_start, x[1]-before_start) + x[2:]
        if before_start and after_start:
            for x in X: yield (x[1]-after_start,  x[1]-before_start) + x[2:]
        if after_end and before_end:
            for x in X: yield (x[0]-after_end,    x[0]-before_end) + x[2:]
        if before_start and after_end:
            for x in X: yield (x[0]-after_end,    x[1]-before_start) + x[2:]

def generate(X, l, **kwargs):
    return bounded(neighborhood(X, l, **kwargs), 0, l)



"""
This module implements the in-memory serialization.
"""

# Built-in modules #
import collections

# Internal modules #
from track.serialize import Serializer

################################################################################
class SerializerMemory(Serializer):
    def __init__(self):
        self.tracks = []

    def newTrack(self, info=None, name=None):
        self.tracks.append(collections.defaultdict(list))

    def newFeature(self, chrom, feature):
        self.tracks[-1]['chrom'].append(feature)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



# coding: utf-8

################################### Modules ####################################
import sys
from track import common
from track.test import samples
from operator import itemgetter

################################# Description ##################################
label           = 'concatenate'
short_name      = 'Concatenate'
long_name       = 'Concatenate N tracks together'
input_tracks    = [{'key':'n_tracks', 'position':1, 'kind':'many', 'fields':['start','end','...']}]
input_args      = []
input_meta      = []
output_tracks   = [{'position':1, 'fields': ['start','end','...']}]
tracks_collapse = None
fields_collapse = 'union'
chroms_collapse = 'union'

################################ Documentation #################################
tooltip = \
"""
The ``concatenate`` manipulation takes only one track
for input. The output consists of all intervals that
were not covered by a feature in the input track.
"""

tickets = []

################################## Examples ####################################
numeric_example = \
"""
X [start,end]: (0,20)
Y [start,end]: (10,30)
R [start,end]: (0,20) (10,30)
"""

visual_example = \
u"""
X: ───▤▤▤▤▤▤▤▤▤────────────▤▤▤▤▤▤▤──────
Y: ──────▤▤▤▤▤────▤▤▤▤▤─────────────────
R: ───▤▤▤▤▤▤▤▤▤───▤▤▤▤▤────▤▤▤▤▤▤▤──────
         ▤▤▤▤▤
"""

#################################### Tests #####################################
tests = [
    {'tracks':   {'n_tracks': [iter([(0,10),(0,10),(5,10)]),
                               iter([(0,5),(0,12),(5,8)])]},
     'expected': [(0,5),(0,10),(0,10),(0,12),(5,8),(5,10)]}
    ,
    {'tracks':   {'n_tracks': [samples['small_features'][1]['sql'],
                               samples['small_features'][4]['sql']]},
     'chromosome': 'chrI',
     'expected': [( 0,   10, u'Validation feature 1',  10.0, 0),
                  ( 2,    8, u'Validation feature 2',   0.0, 0),
                  (10,   20, u'Lorem',                  1.0, 1),
                  (20,   30, u'Validation feature 3',  10.0, 0),
                  (25,   30, u'Validation feature 4',   0.0, 0),
                  (30,   40, u'Ipsum',                  2.0, 1),
                  (40,   45, u'Validation feature 5',   0.0, 0),
                  (40,   50, u'Validation feature 6',  10.0, 0),
                  (60,   70, u'Validation feature 7',  10.0, 0),
                  (70,   80, u'Validation feature 8',  10.0, 0),
                  (90,  100, u'Validation feature 9',  10.0, 0),
                  (90,  110, u'Validation feature 10', 10.0, 0),
                  (120, 130, u'Validation feature 11', 10.0, 0),
                  (125, 135, u'Validation feature 12',  5.0, 0)]}]

############################### Implementation #################################
def generate(n_tracks):
    # Get all iterators #
    sentinel = (sys.maxint, sys.maxint, 0.0)
    tracks   = [common.sentinelize(x, sentinel) for x in n_tracks]
    features = range(len(tracks))
    # Advance feature #
    def advance(i):
        features[i] = tracks[i].next()
        if features[i] == sentinel:
            tracks.pop(i)
            features.pop(i)
    # Find lowest feature #
    def get_lowest_feature():
        i = min(enumerate([(f[0],f[1]) for f in features]), key=itemgetter(1))[0]
        return i, features[i]
    # Core loop #
    for i in xrange(len(tracks)-1, -1, -1): advance(i)
    while tracks:
        i,f = get_lowest_feature()
        yield f
        advance(i)



"""
Provides unittesting to the pacakge. Every format implemented should have its corresponding tests.

To run all the unitests, in the distribution directory, type::

   $ nosetests --with-doctest

The __init__ file here creates a large dictionary of different tracks found in the "samples" directory constisting mainly small validation tracks
This dictionariy is then used by the unittests themselves.
"""

# Built-in modules #
import os

# Internal modules #
from track import formats

################################################################################
# Tracks path #
try:
    project_path = os.path.abspath('/'.join(os.path.realpath(__file__).split('/')[:-1]) + '../../../') + '/'
    samples_path = project_path + 'samples/'
except NameError:
    samples_path = 'samples/'

# Overide with an environement variable #
if os.environ.has_key('SAMPLE_TRACKS_PATH'):
    samples_path = os.environ['SAMPLE_TRACKS_PATH']

# Ending slash #
if not samples_path.endswith('/'): samples_path += '/'

# Chromosome files #
chr_files = {
    'yeast': samples_path + 'chrmeta/sacCer2.chr',
    'small': samples_path + 'chrmeta/small.chr'
}

# Tracks collection #
samples = {
  'small_features': {
    1:            {'name':'Validation features 1', 'filename': 'features1'},
    2:            {'name':'Validation features 2', 'filename': 'features2'},
    3:            {'name':'Validation features 3', 'filename': 'features3'},
    4:            {'name':'Validation features 4', 'filename': 'features4'},
    5:            {'name':'Validation features 5', 'filename': 'features5'},
  },
  'small_signals': {
    1:            {'name':'Signal track 1',        'filename': 'scores1'},
    2:            {'name':'Signal track 2',        'filename': 'scores2'},
    3:            {'name':'Signal track 3',        'filename': 'scores3'},
    4:            {'name':'Signal track 4',        'filename': 'scores4'},
    5:            {'name':'Signal track 5',        'filename': 'scores5'},
    6:            {'name':'Signal track 6',        'filename': 'scores6'},
    7:            {'name':'Signal track 7',        'filename': 'scores7'},
  },
  'rand_features': {
    1:            {'name':'Random features 1',     'filename': 'random1'},
    2:            {'name':'Random features 2',     'filename': 'random2'},
    3:            {'name':'Random features 3',     'filename': 'random3'},
    4:            {'name':'Random features 4',     'filename': 'random4'},
  },
  'rand_signals': {
    'Pol2':       {'name':'Pol2 signal',           'filename': 'yeast_pol2'},
    'Rap1':       {'name':'Rap1 signal',           'filename': 'yeast_rap1'},
  },
  'yeast_features': {
    'All':        {'name':'All yeast genes',       'filename': 'yeast_genes'},
    'Ribi':       {'name':'Yeast Ribi genes',      'filename': 'yeast_ribi_genes'},
    'RP':         {'name':'Yeast RP genes',        'filename': 'yeast_rp_genes'},
  },
  'gff_tracks': {
    1:            {'name':'UCSC GFF example',      'filename': 'gff_test1'},
  },
  'sga_tracks': {
    1:            {'name':'SGA example',           'filename': 'sga_test1'},
  },
  'gtf_tracks': {
    1:            {'name':'WUSTL GTF example',     'filename': 'gtf_test1'},
    'GenRep':     {'name':'GenRep saccer GTF',     'filename': 'gtf_saccer'},
  },
  'gzip_tracks': {
    1:            {'name':'Validation features 1', 'filename': 'features1.bed'},
    4:            {'name':'Validation features 4', 'filename': 'features4'},
  },

}

# Add the path for every format #
for group_key, group in sorted(samples.items()):
    for track_key, track in sorted(group.items()):
        for format in formats:
            track[format] = samples_path + format + '/' + track['filename'] + '.' + format

# Per format lists #
bed_samples = samples['small_features'].values() + \
              samples['yeast_features'].values()

# Specially hard to parse tracks for testing #
if os.path.exists(samples_path):
    formats = ['bed', 'wig']
    outcomes = ['pass', 'fail']
    challanges = {}
    for format in formats:
        challanges[format] = {}
        for outcome in outcomes:
            directory = samples_path + format + '/' + 'should_' + outcome + '/'
            challanges[format][outcome] = [directory + file for file in os.listdir(directory) if file.endswith('.' + format)]

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the parsing of WIG files according to this standard:

http://genome.ucsc.edu/goldenPath/help/wiggle.html
"""

# Built-in modules #
import shlex

# Internal modules #
from track.parse import Parser
from track.common import iterate_lines
from track.util import floats_eq, overlapping

# Constants #
all_fields = ['start', 'end', 'score']

################################################################################
class ParserWIG(Parser):
    format = 'wig'
    def parse(self):
        # Initial variables #
        info   = {}
        params = {}
        declare_track = True
        last_feature  = None
        last_chrom    = None
        # Line loop #
        for number, line in iterate_lines(self.path):
            # Ignored lines #
            if line.startswith("browser "): continue
            # Track headers #
            if line.startswith("track "):
                try:
                    info = dict([p.split('=',1) for p in shlex.split(line[6:])])
                except ValueError:
                    self.handler.error("The track%s seems to have an invalid <track> header line", self.path, number)
                declare_track = True
                continue
            # Have we started a track already ? #
            if declare_track:
                declare_track = False
                if last_feature:
                    self.handler.newFeature(last_chrom, last_feature)
                    last_feature = None
                    last_chrom   = None
                self.handler.newTrack(info, self.name)
                self.handler.defineFields(all_fields)
            # Directive line #
            if line.startswith("variableStep") or line.startswith("fixedStep"):
                params = dict([p.split('=',1) for p in shlex.split('mode=' + line)])
                if not params.get('chrom', False):
                    self.handler.error("The track%s doesn't specify a chromosome.", self.path, number)
                try:
                    params['span'] = int(params.get('span', 1))
                except ValueError:
                    self.handler.error("The track%s has a non integer as span value.", self.path, number)
                if params['span'] < 1:
                    self.handler.error("The track%s has a negative or null span value.", self.path, number)
                if line.startswith("fixedStep "):
                    if not 'start' in params:
                        self.handler.error("The track%s has a fixedStep directive without a start.", self.path, number)
                    try:
                        params['start'] = int(params['start'])
                    except ValueError:
                        self.handler.error("The track%s has a non integer as start value.", self.path, number)
                    try:
                        params['step'] = int(params.get('step',1))
                    except ValueError:
                        self.handler.error("The track%s has a non integer as step value.", self.path, number)
                    if params['step'] < 1:
                        self.handler.error("The track%s has a negative or null step value.", self.path, number)
                continue
            # Not a directive line #
            if not params:
                self.handler.error("The track%s is missing a fixedStep or variableStep directive.", self.path, number)
            # Fixed #
            if params['mode'] == 'fixedStep':
                try:
                    line = float(line)
                except ValueError:
                    self.handler.error("The track%s has non floats as score values.", self.path, number)
                chrom   = params['chrom']
                feature = [params['start'], params['start'] + params['span'], line]
                params['start'] += params['span'] + params['step'] - 1
            # Variable #
            elif params['mode'] == 'variableStep':
                line = line.split('\t')
                if len(line) == 1: line = line[0].split()
                try:
                    line[0] = int(line[0])
                    line[1] = float(line[1])
                except ValueError:
                    self.handler.error("The track%s has invalid values.", self.path, number)
                except IndexError:
                    self.handler.error("The track%s has missing values.", self.path, number)
                chrom   = params['chrom']
                feature = [line[0], line[0] + params['span'], line[1]]
            # Ignore null scores #
            if feature[2] == 0.0: continue
            # Merge adjacent features with same scores #
            # For instance ['chr1', 10, 11, 9.8] and ['chr1', 11, 12, 9.8] should merge.
            if last_feature:
                if last_chrom == chrom:
                    if last_feature[1] > feature[0]:
                        self.handler.error("The track%s has a start or span larger than its end or step.", self.path, number)
                    if floats_eq(last_feature[2], feature[2]) and overlapping(last_feature[0], last_feature[1], feature[0], feature[1]):
                        last_feature[0] = min(last_feature[0], feature[0])
                        last_feature[1] = max(last_feature[1], feature[1])
                        continue
                self.handler.newFeature(last_chrom, last_feature)
            last_feature = feature
            last_chrom   = chrom
        if last_feature: self.handler.newFeature(last_chrom, last_feature)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains tests for the BED format.
We will do a roundtrips for several sample validation tracks.
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.common import temporary_path, assert_file_equal
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestConversion(unittest.TestCase):
    def runTest(self):
        for num, info in sorted(samples['gtf_tracks'].items()):
            # This one is too large #
            if num == 'GenRep': continue
            # Prepare paths #
            orig_gtf_path = info['gtf']
            orig_sql_path = info['sql']
            test_sql_path = temporary_path('.sql')
            # From BED to SQL #
            track.convert(orig_gtf_path, test_sql_path, assembly='sacCer2')
            self.assertTrue(assert_file_equal(orig_sql_path, test_sql_path))
            # Clean up #
            os.remove(test_sql_path)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Run all the manips unittest by creating a test suite.

To run this in ipython type::

    [1] from track.manipulate import tests, list_of_manip_names
    [2] t = tests.TestManips()
    [3] for name in list_of_manip_names: print name, getattr(t,name)()
"""

# Built-in modules #
import os

# Internal modules #
import track
from track import manipulate
from track.manipulate import window_smoothing, mean_score_by_feature
from track.manipulate import complement, merge_scores, threshold
from track.common import temporary_path
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestManips(unittest.TestCase):
    def runTest():pass

def make_fn(manip_name):
    def test_fn(self):
        self.maxDiff = None
        manip = getattr(manipulate, manip_name)
        for got, expected in manip.test(): self.assertEqual(got, expected)
    return test_fn

for name in manipulate.list_of_manip_names:
    setattr(TestManips, name, make_fn(name))
suite = unittest.TestLoader().loadTestsFromTestCase(TestManips)

###################################################################################
class TestComplement(unittest.TestCase):
    """Test a complement call via tracks."""
    def runTest(self):
        in_path = samples['small_features'][1]['sql']
        out_path = temporary_path('.bed')
        with track.load(in_path) as t:
            result = complement(t)
            result.export(out_path)
        with track.load(out_path) as t:
            data = t.read('chrI')
            got = map(tuple, data)
        expected = [( 10,     20),
                    ( 30,     40),
                    ( 50,     60),
                    ( 80,     90),
                    (110,    120),
                    (135, 230208)]
        self.assertEqual(got, expected)
        os.remove(out_path)

class TestWindowSmooth(unittest.TestCase):
    """Test a window_smoothing call via tracks."""
    def runTest(self):
        in_path = temporary_path('.sql')
        out_path = temporary_path('.sql')
        with track.new(in_path) as t:
            t.fields = ('start','end','score')
            t.assembly = 'sacCer2'
            t.write('chrI',[(0,2,10),(2,4,20),(6,8,10)])
            result = window_smoothing(t, 2)
            result.export(out_path)
        with track.load(out_path) as t:
            data = t.read('chrI')
            got = map(tuple, data)
        expected = [(0,  1,   8.0),
                    (1,  3,  12.0),
                    (3,  5,  10.0),
                    (5,  6,   8.0),
                    (6,  9,   4.0),
                    (9,  10,  2.0)]
        self.assertEqual(got, expected)
        os.remove(in_path)
        os.remove(out_path)

###################################################################################
class TestMergeScores(unittest.TestCase):
    """Test a merge scores call via files."""
    def runTest(self):
        in_paths = [samples['small_signals'][1]['sql'],
                    samples['small_signals'][2]['sql'],
                    samples['small_signals'][3]['sql']]
        out_path = temporary_path('.sql')
        t = merge_scores(in_paths)
        t.export(out_path)
        with track.load(out_path) as t:
            data = t.read('chrI')
            got = map(tuple, data)
        expected = [( 0,    5,    2.0 + 0.6666666666666666),
                    ( 5,   10,    4.0),
                    ( 20,  30,   10.0),
                    ( 30,  40,   30.0),
                    ( 40,  50,   26.0 + 0.666666666666666),
                    ( 50,  60,  120.0),
                    ( 60,  68,  100.0),
                    ( 68,  70,  200.0),
                    ( 70,  80,  100.0),
                    ( 90, 110,    3.0),
                    (120, 130,   10.0)]
        self.assertEqual(got, expected)
        os.remove(out_path)

class TestMeanScores(unittest.TestCase):
    """Test a mean score by feature call via files."""
    def runTest(self):
        x_path = samples['small_signals'][4]['sql']
        y_path = samples['small_features'][4]['sql']
        out_path = temporary_path('.sql')
        t = mean_score_by_feature(x_path,y_path)
        t.export(out_path)
        with track.load(out_path) as t:
            data = t.read('chrI')
            got = map(tuple, data)
        expected = [(10, 20, 15.0, u'Lorem', 1),
                    (30, 40, 50.0, u'Ipsum', 1)]
        self.assertEqual(got, expected)
        os.remove(out_path)

class TestThreshold(unittest.TestCase):
    """Test a threshold call via files."""
    def runTest(self):
        in_path = samples['small_signals'][4]['sql']
        out_path = temporary_path('.sql')
        t = threshold(in_path, 8000.0)
        t.export(out_path)
        with track.load(out_path) as t:
            data = t.read('chrI')
            got = map(tuple, data)
        expected = [(120, 122, 9000.0)]
        self.assertEqual(got, expected)
        os.remove(out_path)



"""
Usefull stuff for the track package.
"""

# Built-in modules #
import os, shlex

# Internal modules #
from track.common import temporary_path, iterate_lines

###############################################################################
# Constants #
sql_field_types = {'start':        'integer',
                   'end':          'integer',
                   'score':        'real',
                   'strand':       'integer',
                   'name':         'text',
                   'thick_start':  'integer',
                   'thick_end':    'integer',
                   'item_rgb':     'text',
                   'block_count':  'integer',
                   'block_sizes':  'text',
                   'block_starts': 'text',
                   'attributes':   'text',
                   'source':       'text',
                   'frame':        'integer',
                   'length':       'integer',
                   'attributes':   'text',}

py_field_types  = {'start':        int,
                   'end':          int,
                   'score':        float,
                   'strand':       int,
                   'name':         str,
                   'thick_start':  int,
                   'thick_end':    int,
                   'item_rgb':     str,
                   'block_count':  int,
                   'block_sizes':  str,
                   'block_starts': str,
                   'attributes':   str,
                   'source':       str,
                   'frame':        int,
                   'length':       int,
                   'attributes':   str,}

format_synonyms = {'db': 'sql',
                   'bw': 'bigwig',
                   'bwg': 'bigwig',
                   'wiggle_0': 'wig',}

###############################################################################
def determine_format(path):
    """Try to guess the format of a track given its path.
       Returns a three letter extension."""
    # Get the extension #
    ext = os.path.splitext(path)[1][1:].lower()
    # An extension is provided #
    if ext: return format_synonyms.get(ext, ext)
    # Does the file exist #
    if not os.path.exists(path):
        raise Exception("The format of the path '%s' cannot be determined. Please specify a format or add an extension." % path)
    # Try our own sniffing now #
    with open(path, 'r') as file:
        file_format = guess_file_format(file)
    # If still nothing, raise exception #
    if not file_format:
        raise Exception("The format of the file '%s' cannot be determined. Please specify a format or add an extension." % path)
    # Return the format #
    file_format = file_format.lower()
    return format_synonyms.get(file_format, file_format)

###############################################################################
def guess_file_format(handle):
    """Try to guess the format of a track given its content.
       Returns a three letter extension."""
    # Check SQLite #
    if handle.read(15) == "SQLite format 3": return 'sql'
    handle.seek(0)
    # Try to read the track line #
    for number, line in enumerate(handle):
        line = line.strip("\n").lstrip()
        if number > 100: break
        if line.startswith("track "):
            try: id = dict([p.split('=',1) for p in shlex.split(line[6:])]).get('type', '')
            except ValueError: id = ''
            return id

###############################################################################
def gzip_inner_format(path):
    """Try to guess the format of a file inside a
       compressed gzip archive. Returns a three
       letter extension"""
    # Get the extension #
    ext = os.path.splitext(path.strip('.gz').strip('.gzip'))[1][1:]
    # An extension is provided #
    if ext: return format_synonyms.get(ext, ext)
    # Try our own sniffing now #
    import gzip
    file = gzip.open(path, 'r')
    file_format = guess_file_format(file)
    file.close()
    # If still nothing, raise exception #
    if not file_format:
        raise Exception("The inner format of the gzip file '%s' cannot be determined. Please specify a format or add an extension." % path)
    # Return the format #
    return format_synonyms.get(file_format, file_format)

###############################################################################
def parse_chr_file(self, path):
    """Read a chromsome file and return a dictionary"""
    chrmeta = {}
    for number, line in iterate_lines(path):
        items = line.split('\t')
        if len(items) == 1: items = line.split()
        if len(items) != 2:
            raise Exception("The file '" + path + ":" + str(number) + "' does not seam to be a valid chromosome file.")
        name = items[0]
        try:
            length = int(items[1])
        except ValueError:
            raise Exception("The file '" + path + ":" + str(number) + "' has non-integer as chromosome lengths.")
        chrmeta[name] = dict([('length', length)])
    if not chrmeta:
        raise Exception("The file '" + path + "' does not seam to contain any information.")
    return chrmeta

###############################################################################
def serialize_chr_file(chrmeta, path=None, seperator='\t'):
    """Read a dictionary and create a plain text file"""
    if not path: path = temporary_path()
    def lines():
        for k,v in chrmeta.items(): yield k + seperator + str(v['length']) + '\n'
    with open(path, 'w') as f: f.writelines(lines())
    return path

###############################################################################
def parse_header_line(handle):
    """
    Tries to parse the 'track' header line and returns a dictionary.
    For instance for the following line:
        track type=wiggle_0 name="DNA Footprints" source="University of Washington"
    You will get the following dictionary:
        {'name': 'DNA Footprints', 'source': 'University of Washington', 'type': 'wiggle_0'}
    """
    result = {}
    for line in handle:
        line = line.strip()
        if len(line) == 0:              continue
        if line.startswith("#"):        continue
        if line.startswith("browser "): continue
        if line.startswith("track "):
            try: result = dict([p.split('=',1) for p in shlex.split(line[6:])])
            except ValueError: pass
        break
    handle.seek(0)
    return result

################################################################################
def add_chromsome_prefix(sel, data):
    """Add the chromosome in front of every feature"""
    if type(sel) == str: chrom = (sel,)
    else:                chrom = (sel['chr'],)
    for f in data: yield chrom + tuple(f)

################################################################################
def join_read_queries(track, selections, fields):
    """Join read results when selection is a list"""
    for sel in selections:
        for f in add_chromsome_prefix(sel, track.read(sel, fields)): yield f

################################################################################
def make_cond_from_sel(selection):
    """Make an SQL condition string from a selection dictionary"""
    query = ""
    # Case start #
    if "start" in selection:
        query += "end > " + str(selection['start'])
        if selection.get('inclusion') == 'strict': query += " and start >= " + str(selection['start'])
    # Case end #
    if "end" in selection:
        if query: query += " and "
        query += "start < " + str(selection['end'])
        if selection.get('inclusion') == 'strict': query += " and end <= " + str(selection['end'])
    # Case strand #
    if "strand" in selection:
        if query: query += " and "
        query += 'strand == ' + str(selection['strand'])
    # Case score #
    if "score" in selection:
        if not isinstance(selection['score'], tuple): raise Exception("Score intervals must be tuples of size 2")
        if query: query += " and "
        query += 'score >= ' + str(selection['score'][0]) + ' and score <= ' + str(selection['score'][1])
    return query

################################################################################
def strand_to_int(strand):
    if strand == '+': return 1
    if strand == '-': return -1
    return 0

def int_to_strand(num):
    num = int(num)
    if num == 1: return  '+'
    if num == -1: return '-'
    return '.'

################################################################################
def floats_eq(a,b, epsilon=0.000001):
    """
    Return true if two floats are equals
    """
    return abs(a-b) < epsilon

def overlapping(start1, stop1, start2, stop2):
    """
    Return true if the two features 1 & 2 are overlapping.
    """
    return start1 <= stop2 and stop1 >= start2

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This directory contains all manipulations.
One manipulation per file.
The file describes what is needed for the manipulation to run.
Things such as the number of tracks needed and extra parameters.
"""
